""" Test around automatic command generation"""

import omnibenchmark.utils.auto_command as omni
from omnibenchmark.utils.exceptions import InterpreterError
import pytest

### Test get_interpreter_from_extension
def test_get_interpreter_from_extension_existing():
    assert omni.get_interpreter_from_extension("sh") == "bash"


def test_get_interpreter_from_extension_lower():
    assert omni.get_interpreter_from_extension("Py") == "python"


def test_get_interpreter_from_extension_remove_dot():
    assert omni.get_interpreter_from_extension(".R") == "R"


def test_get_interpreter_from_extension_unknown():
    with pytest.raises(
        InterpreterError,
        match=r"Interpreter could not be identified from extention xy.*?",
    ):
        omni.get_interpreter_from_extension("xy")


### Test get_command_start_from_interpreter
def test_get_command_start_from_interpreter_existing():
    assert omni.get_command_start_from_interpreter("R") == "Rscript --vanilla"


def test_get_command_start_from_interpreter_unknown():
    with pytest.raises(
        InterpreterError,
        match=r"Command start could not be identified from interpreter fancy.*?",
    ):
        omni.get_command_start_from_interpreter("fancy")


### Test parse_command_line_args
def test_parse_command_line_args_existing():
    assert omni.parse_command_line_args({"input1": "in1", "input2": "in2"}) == [
        '--input1 "in1"',
        '--input2 "in2"',
    ]


### Test automatic_command_generation
def test_automatic_command_generation_with_interpretor():
    command = omni.automatic_command_generation(
        script="path/to/script.py",
        interpreter="python",
        inputs={"input1": "in1", "input2": "in2"},
    )
    assert command == 'python path/to/script.py --input1 "in1" --input2 "in2"'


def test_automatic_command_generation_no_interpretor():
    command = omni.automatic_command_generation(
        script="path/to/script.py", inputs={"input1": "in1", "input2": "in2"}
    )
    assert command == 'python path/to/script.py --input1 "in1" --input2 "in2"'


def test_automatic_command_generation_all():
    command = omni.automatic_command_generation(
        script="path/to/script.py",
        inputs={"input1": "in1", "input2": "in2"},
        parameters={"param1": "p1"},
        outputs={"output1": "out1"},
    )
    assert (
        command
        == 'python path/to/script.py --input1 "in1" --input2 "in2" --output1 "out1" --param1 "p1"'
    )


def test_automatic_command_generation_none():
    command = omni.automatic_command_generation(script="path/to/script.py")
    assert command == "python path/to/script.py"



"""General decorators used in omnibenchmark"""

import functools
from typing import Mapping


def option_list(function):
    """
    Return an empty list, if no or only None type arguments are provided
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        arg_list = list(args) + list(kwargs.values())
        if len(arg_list) > 0 and any(argv is not None for argv in arg_list):
            return function(*args, **kwargs)
        else:
            return []

    return wrapper


def option_str(function):
    """
    Return an empty string, if no or only None type arguments are provided
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        arg_list = list(args) + list(kwargs.values())
        if len(arg_list) > 0 and any(argv is not None for argv in arg_list):
            return function(*args, **kwargs)
        else:
            return ""

    return wrapper


def option_dict_list(function):
    """
    Return an empty list, if none of the provided arguments is a dictionary
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        arg_list = list(args) + list(kwargs.values())
        if len(arg_list) > 0 and all(isinstance(argv, Mapping) for argv in arg_list):
            return function(*args, **kwargs)
        else:
            return []

    return wrapper


def option_dict_none(function):
    """
    Return None, if none of the provided arguments is a dictionary
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        arg_list = list(args) + list(kwargs.values())
        if len(arg_list) > 0 and all(isinstance(argv, Mapping) for argv in arg_list):
            return function(*args, **kwargs)
        else:
            return None

    return wrapper



from typing import TypeVar, Mapping, List, TypedDict, Optional
from pathlib import Path

from omnibenchmark.utils.user_input_checks import (
    check_name_matching,
    flatten,
    mixed_flatten,
    check_default_parameter,
    elements_to_list,
)
from omnibenchmark.utils.auto_input import (
    get_input_files_from_prefix,
    get_parameter_from_dataset,
    drop_none_elements,
)
from omnibenchmark.utils.exceptions import InputError
from omnibenchmark.management.data_commands import update_datasets_by_keyword
from omnibenchmark.management.parameter_checks import (
    filter_parameter,
    get_all_parameter_combinations,
    filter_parameter_combinations,
)

PathLike = TypeVar("PathLike", str, Path, None)


class OmniInput:
    """Class to store metadata of datasets and files that are specified as inputs in an omnibenchmark project"""

    def __init__(
        self,
        names: List[str],
        prefix: Optional[Mapping[str, List[str]]] = None,
        input_files: Optional[Mapping[str, Mapping[str, str]]] = None,
        keyword: Optional[List[str]] = None,
        default: Optional[str] = None,
        filter_names: Optional[List[str]] = None,
        filter_duplicated: bool = True,
    ):
        """Class to manage inputs of an omnibenchmark

        Args:
            names (List[str]): Names of the input filetypes
            prefix (Optional[Mapping[str, List[str]]], optional): Substrings of the input filetypes. Defaults to None.
            input_files (Optional[Mapping[str, Mapping[str, str]]], optional): Input files ordered by file types.
                                                                               Defaults to None.
            keyword (Optional[List[str]], optional): Keyword to define which datasets are imported as input datasets.
                                                     Defaults to None.
            default (Optional[str], optional): Default input name (e.g., dataset). Defaults to None.
            filter_names (Optional[List[str]], optional): Input dataset names to be ignored. Defaults to None.
        """
        self.names = names
        self.prefix = prefix
        self.input_files = input_files
        self.keyword = keyword
        self.default = default
        self.filter_names = filter_names
        self.filter_duplicated = filter_duplicated

        if self.input_files is None:

            if self.prefix is None or self.keyword is None:
                raise InputError(
                    "Missing input file definition. \n"
                    "Please specify explicit input files or \n"
                    "define prefixes AND dataset keyword(s)to automatically identify them."
                )

            check_name_matching(self.names, self.prefix.keys())
            self.input_files = get_input_files_from_prefix(
                self.prefix,
                self.keyword,
                self.filter_names,
                filter_duplicated=self.filter_duplicated,
            )

        if len(self.input_files) < 1:
            self.input_files = None
            print(
                f"WARNING: No input files in the current project detected.\n"
                f"Run OmniObject.update_object() to update and import input datasets.\n"
                f"Otherwise check the specified prefixes: {self.prefix}"
            )

        if self.input_files is not None:
            check_name_matching(
                self.names, flatten([pre.keys() for pre in self.input_files.values()])
            )

            if self.default is None:
                self.default = next(iter(self.input_files.items()))[0]

            if self.default not in self.input_files.keys():
                raise InputError(
                    f"Input default {self.default} not found. Must have corresponding input files."
                )

            self.default_files = self.input_files[self.default]

    def update_inputs(
        self,
        orchestrator: str,
        query_url: str,
        data_url: str,
        gitlab_url: str,
        check_o_url: bool = True,
        n_latest: int = 9,
        all: bool = True,
    ):
        """Update datasets and files that belong to this OmniInput object.
           This will also import new Datasets with the specified keyword..

        Args:
            orchestrator (str):  Orchestrator url that links all valid projects.
            query_url (str): URL to the knowledgebase dataset query API.
            data_url (str): URL to the knowledgebase dataset API.
            gitlab_url (str): General Gitlab url.
        """
        if self.keyword is not None:
            for key in self.keyword:
                update_datasets_by_keyword(
                    keyword=key,
                    filter_names=self.filter_names,
                    o_url=orchestrator,
                    query_url=query_url,
                    data_url=data_url,
                    gitlab_url=gitlab_url,
                    check_o_url=check_o_url,
                    n_latest=n_latest,
                    all=all,
                )
            if self.prefix is not None:
                check_name_matching(self.names, self.prefix.keys())
                self.input_files = get_input_files_from_prefix(
                    self.prefix,
                    self.keyword,
                    self.filter_names,
                    filter_duplicated=self.filter_duplicated,
                )
                check_name_matching(
                    self.names,
                    flatten([pre.keys() for pre in self.input_files.values()]),
                )
            if self.input_files is not None and len(self.input_files) > 0:
                if self.default is None or self.default not in self.input_files.keys():
                    self.default = next(iter(self.input_files.items()))[0]


class OmniParameter:
    """Class to store metadata of datasets and files that are specified as parameter in an omnibenchmark project"""

    def __init__(
        self,
        names: List[str],
        values: Optional[Mapping[str, List]] = None,
        default: Optional[Mapping[str, str]] = None,
        keyword: Optional[List[str]] = None,
        filter: Optional[Mapping[str, str]] = None,
        combinations: Optional[List[Mapping[str, str]]] = None,
    ):
        """Initiate an instance of the class OmniParameter

        Args:
            names (List[str]): Name of all valid parameter
            values (Optional[Mapping[str, List]], optional): Parameter values, automatically detected from parameter dataset.
                                                             Defaults to None.
            default (Optional[Mapping[str, str]], optional): Default parameter values. Defaults to None.
            keyword (Optional[List[str]], optional): Keyword to import the parameter dataset with. Defaults to None.
            filter (Optional[Mapping[str, str]], optional): Filter to use for the parameter space. Defaults to None.
            combinations (Optional[List[Mapping[str, str]]], optional): All possible parameter combinations. Defaults to None.
        """
        self.names = names
        self.values = values
        self.default = default
        self.keyword = keyword
        self.filter = filter
        self.combinations = combinations

        if self.values is None and self.keyword is not None:
            self.values = get_parameter_from_dataset(self.names, self.keyword)

        if self.values is not None:
            check_name_matching(self.names, self.values.keys())
            self.values = elements_to_list(self.values)
            self.values = filter_parameter(self.values, self.filter)

            if self.combinations is None:
                self.combinations = get_all_parameter_combinations(self.values)

        if self.combinations is not None and len(self.combinations) > 0:
            check_name_matching(
                self.names, flatten([comb.keys() for comb in self.combinations])
            )
            self.combinations = filter_parameter_combinations(
                self.combinations, self.filter
            )
            if self.default is None or len(self.default) == 0:
                self.default = self.combinations[0]

        self.default = check_default_parameter(self.default, self.combinations)

    def update_parameter(
        self,
        orchestrator: str,
        query_url: str,
        data_url: str,
        gitlab_url: str,
        check_o_url: bool = True,
        n_latest: int = 9,
    ):
        """Update datasets and files that belong to this OmniParameter object.
           This will also import new Datasets with the specified keyword.
        Args:
            orchestrator (str): Orchestrator url that links all valid projects.
            query_url (str): URL to the knowledgebase dataset query API.
            data_url (str): URL to the knowledgebase dataset API.
            gitlab_url (str): General Gitlab url.
            check_o_url (bool): If inclusion to an orchestrator should be checked.
        """
        if self.keyword is not None:
            for key in self.keyword:
                update_datasets_by_keyword(
                    keyword=key,
                    o_url=orchestrator,
                    query_url=query_url,
                    data_url=data_url,
                    gitlab_url=gitlab_url,
                    check_o_url=check_o_url,
                    n_latest=n_latest,
                )
            if self.values is None:
                self.values = {}
            val = get_parameter_from_dataset(self.names, self.keyword)
            self.values = {
                x: mixed_flatten(drop_none_elements([val.get(x), self.values.get(x)]))
                for x in set(self.values).union(val)
            }
            self.values = elements_to_list(self.values)
            check_name_matching(self.names, self.values.keys())
            self.values = filter_parameter(self.values, self.filter)
            self.combinations = get_all_parameter_combinations(self.values)

            if self.combinations is not None:
                check_name_matching(
                    self.names, flatten([comb.keys() for comb in self.combinations])
                )
                self.combinations = filter_parameter_combinations(
                    self.combinations, self.filter
                )
                if self.default is None or len(self.default) == 0:
                    self.default = self.combinations[0]

            self.default = check_default_parameter(self.default, self.combinations)


class OutMapping(TypedDict):
    output_files: Optional[Mapping]
    input_files: Optional[Mapping]
    parameter: Optional[Mapping]



"""Commands related to import and update relevant datasets"""
from typing import List, Mapping, Any, Optional, Tuple
from renku.api import Dataset
from omnibenchmark.renku_commands import renku_api
from omnibenchmark.renku_commands.datasets import (
    renku_dataset_import,
    renku_dataset_update,
    renku_add_to_dataset,
    renku_unlink_from_dataset,
    renku_dataset_remove,
)
from omnibenchmark.utils.exceptions import InputError
from omnibenchmark.management.data_checks import query_multipages
import requests
import os
import urllib
import gitlab
from omnibenchmark.renku_commands.general import renku_save
from iteration_utilities import unique_everseen  # type: ignore


# Find datasets by string
def query_datasets_by_string(
    string: str,
    url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    page_item: int = 100,
) -> List[Mapping[Any, Any]]:
    """Query datasets in the knowledge base by a string.

    Args:
        string (str): String to query datasets for
        url (str): URL to the knowledgebase dataset query API. Default "https://renkulab.io/knowledge-graph/datasets?query=".

    Returns:
        List[Mapping[Any, Any]]: List of all datasets associated to that string with their metadata
    """
    query_url = url + string
    return query_multipages(url=query_url, page_item=page_item)


# Find dataset by match of property
def query_datasets_by_property(
    string: str,
    url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    match_string: Optional[str] = None,
    property_name: str = "keywords",
) -> List[Mapping[Any, Any]]:
    """Query datasets in a knowledge base by string and matching property.

    Args:
        string (str): String to query datasets for
        url (str): URL to the knowledgebase dataset query API. Default "https://renkulab.io/knowledge-graph/datasets?query=".
        match_string (Optional[str], optional): String to filter matching properties by.
                                                Will be the same as the query string, if None.
                                                Defaults to None.
        property_name (str, optional): Property to filter datasets by. Defaults to "keywords".

    Returns:
        List[Mapping[Any, Any]]: List of all datasets associated to that string
                                 with the property matching the "match_string" with their metadata.
    """
    if match_string is None:
        match_string = string
    all_data_json = query_datasets_by_string(string=string, url=url)
    matched_dataset: List = []
    for dataset in all_data_json:
        data_prop = (
            [dataset[property_name]]
            if not isinstance(dataset[property_name], list)
            else dataset[property_name]
        )
        if any(prop_val.lower() == match_string.lower() for prop_val in data_prop):
            matched_dataset.append(dataset)
    return matched_dataset


# Filter exisiting datasets
def filter_existing(data_json: List) -> Tuple[List, List]:
    """Split list of datasets into a list of thse that already exist in the project and those that do not exist by name.

    Args:
        data_json (List): List of datasets with their metadata mappings

    Returns:
        Tuple[List, List]: First List contains not existing datasets,
                           second list existing datasets in the current project
    """
    update_list: List = []
    import_list: List = []
    # datasets = Dataset.list()
    datasets = renku_api.renku_dataset_list()
    name_list = [dataset.name for dataset in datasets]
    for data in data_json:
        if data["name"] in name_list:
            update_list.append(data)
        else:
            import_list.append(data)
    return import_list, update_list


# Apply predefined filter
def import_filter(data_json: List, filter_names: Optional[List[str]] = None) -> List:
    """Filter import datasets by specified names.

    Args:
        data_json (List): List of datasets with their metadata mappings
        filter_names (Optional[List[str]], optional): Names to be filtered from the dataset list. Defaults to None.

    Returns:
        List: Filtered list of dataset metadata
    """
    if filter_names is None:
        return data_json
    else:
        return [data for data in data_json if data["name"] not in filter_names]


# Get dataset field by matching property
def get_field_by_dataset_property(
    string: str,
    url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    property_name: str = "keywords",
    field: str = "identifier",
    filter_ex: bool = False,
    filter_names: Optional[List[str]] = None,
) -> Tuple[List[str], List[str]]:
    """Get a specific field of all datasets that match a specified query.

    Args:
        string (str): String to query datasets for
        url (str, optional): URL to the knowledgebase dataset query API.
        property_name (str, optional): Dataset property to match with the query string. Defaults to "keywords".
        field (str, optional): Dataset field to return. Defaults to "identifier".
        filter_ex (bool, optional): If true return the name of all existing datasets instead. Defaults to False.
        filter_names (Optional[List[str]], optional): Names to be filtered from the dataset list. Defaults to None.

    Returns:
        Tuple[List[str], List[str]]: List with all or not already existing datasets.
                                     List with names of existing datasets.
    """
    data_json = query_datasets_by_property(
        string=string, url=url, property_name=property_name
    )
    data_json = import_filter(data_json=data_json, filter_names=filter_names)
    up_json: List = []
    if filter_ex:
        data_json, up_json = filter_existing(data_json=data_json)

    return (
        [dataset[field] for dataset in data_json],
        [dataset["name"] for dataset in up_json],
    )


# Get data info by dataset id
def get_data_info_by_id(
    id: str, url: str = "https://renkulab.io/knowledge-graph/datasets/"
) -> Mapping[Any, Any]:
    """Get dataset metadat by dataset id

    Args:
        id (str): Renku dataset identifier
        url (str, optional): URL to the knowledgebase dataset API. Defaults to "https://renkulab.io/knowledge-graph/datasets/".

    Returns:
        Mapping[Any, Any]: Dataset metadata (e.g., name, description, keywords, links, creator, ...)
    """
    query_url = url + id
    response = requests.get(query_url)
    return response.json()


def find_dataset_linked_to_wflow(
    info_list: List[Mapping], top: int = 2, filter_pat: str = "meta"
) -> List[Mapping]:
    """Check a list of datasets for workflows associated to their linked files.

    Args:
        info_list (List[Mapping]): List of datasets and their metadata to filter
        top (int, optional): Number of associated files to check for workflow association Defaults to 2.
        filter_pat (str, optional): Filter pattern to exclude files by. Defaults to "meta".

    Returns:
        List[Mapping]: List of datasets and metadata with files that have a workflow linked to them.
    """
    origin_info = []
    for info in info_list:
        data_files = info["hasPart"]
        data_file_list = [
            data_fi["atLocation"]
            for data_fi in data_files
            if filter_pat not in data_fi["atLocation"]
        ]
        for data_file in data_file_list[:top]:
            file_lineage = requests.get(
                info["project"]["_links"][0]["href"]
                + "/files/"
                + urllib.parse.quote(data_file, safe="")
                + "/lineage"
            )
            if "message" in file_lineage.json().keys():
                continue
            if "edges" in file_lineage.json().keys():
                if any(
                    data_file in fi["target"] for fi in file_lineage.json()["edges"]
                ):
                    if info not in origin_info:
                        origin_info.append(info)
    return origin_info


def filter_duplicated_names(info_list: List[Mapping]) -> List[Mapping]:
    """Filter a list of datasets to get only the original dataset if their are multiple datasets found with the same name.

    Args:
        info_list (List[Mapping]): List of datasets with metadata to filter

    Returns:
        List[Mapping]: Dataset list with unique datasets
    """
    info_list = list(unique_everseen(info_list))
    name_list = [info["name"] for info in info_list]
    dup_names = set(
        [info_name for info_name in name_list if name_list.count(info_name) > 1]
    )
    uni_list = [info for info in info_list if info["name"] not in dup_names]
    for dup in dup_names:
        dup_info = [info for info in info_list if info["name"] == dup]
        dup_project = set([info["project"]["_links"][0]["href"] for info in dup_info])
        if len(dup_project) == 1:
            all_datasets = query_multipages(url=list(dup_project)[0] + "/datasets")
            all_ids = [dat["identifier"] for dat in all_datasets]
            origin_info = [info for info in dup_info if info["identifier"] in all_ids]
        else:
            o_info = find_dataset_linked_to_wflow(dup_info)
            o_project = set([info["project"]["_links"][0]["href"] for info in o_info])
            if len(o_project) == 1:
                all_datasets = query_multipages(url=list(o_project)[0] + "/datasets")
                all_ids = [dat["identifier"] for dat in all_datasets]
                origin_info = [info for info in o_info if info["identifier"] in all_ids]
            else:
                origin_info = []
        if len(origin_info) != 1:
            print(
                f"WARNING: Could not identify origin of {dup}.\n"
                f"If you anyways want to use and import it, consider manual import:\n"
                f"renku dataset import DATA_URL"
            )
            continue
        uni_list.append(*origin_info)
    return uni_list


# Get origin dataset ids
def get_origin_dataset_ids(
    ids: List[str], url: str = "https://renkulab.io/knowledge-graph/datasets/"
) -> List[Mapping]:
    """Filter a list of dataset ids for original dataset ids
       (datasets associated to projects, where they were generated and not imported).

    Args:
        ids (List[str]): List of dataset ids to filter
        url (str, optional): URL to the knowledgebase dataset API. Defaults to "https://renkulab.io/knowledge-graph/datasets/".

    Returns:
        List[Mapping]: Filtered datasets and their metadata
    """
    all_infos: List = []
    for data_id in ids:
        data_info = get_data_info_by_id(id=data_id, url=url)
        if "sameAs" in data_info.keys():
            if not data_info["sameAs"] == data_info["url"]:
                same_id = data_info["sameAs"].split("/")[
                    -1
                ]  # Unstable string manipulation??
                data_info = get_data_info_by_id(id=same_id, url=url)
        if data_info not in all_infos and "name" in data_info.keys():
            all_infos.append(data_info)
    origin_infos = filter_duplicated_names(all_infos)
    return origin_infos


def get_project_info_from_url(project_url: str) -> Mapping[Any, Any]:
    """Get project metadata from url

    Args:
        project_url (str): URL to the knowledgebase dproject URL.

    Returns:
        Mapping[Any, Any]: Project metadata dictionary
    """
    response = requests.get(project_url)
    return response.json()


def check_orchestrator(
    data_info: Mapping,
    o_url: str,
    gitlab_url: str = "https://renkulab.io/gitlab",
    n_latest: int = 9,
) -> Optional[str]:
    """Check if a dataset is associated to a project that is part of the specified orchestrators projects.

    Args:
        data_info (Mapping): Dataset metadata to check
        o_url (str): orchestrator url to check if the dataset project is associated to.
        gitlab_url (url, optional): General Gitlab url. Defaults to "https://renkulab.io/gitlab".

    Returns:
        Optional[str]: Dataset url, if the dataset is associated to a project that is part of the orchestrator.
    """
    project_url = data_info["project"]["_links"][0]["href"]
    project_info = get_project_info_from_url(project_url)
    if "identifier" not in project_info.keys():
        print(
            f"Warning: Can't find dataset at {project_url}.\n"
            f"Please check {data_info['name']}.\n"
        )
        return None
    o_info = get_project_info_from_url(o_url)
    renku_git = gitlab.Gitlab(gitlab_url)
    o_git = renku_git.projects.get(o_info["identifier"])
    success = o_git.pipelines.list(status="success", all=False)
    latest = o_git.pipelines.list(order_by="updated_at", all=False)[
        :n_latest
    ]  # type:ignore
    query_pipes = list(set(success + latest))  # type:ignore
    bridge_projects: List = []
    for pipe in query_pipes:
        bridge_projects.extend(
            [
                brid.downstream_pipeline["project_id"]
                for brid in pipe.bridges.list()
                if brid.status == "success"
            ]
        )
        if project_info["identifier"] in bridge_projects:
            return data_info["url"]
    return None


def get_data_url_by_keyword(
    keyword: str,
    o_url: str,
    filter_ex: bool = False,
    filter_names: Optional[List[str]] = None,
    query_url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    data_url: str = "https://renkulab.io/knowledge-graph/datasets/",
    gitlab_url: str = "https://renkulab.io/gitlab",
    check_o_url: bool = True,
    n_latest: int = 9,
) -> Tuple[List[str], List[str]]:
    """Get all valid dataset urls by matching keywords.

    Args:
        keyword (str): keyword to find datasets by
        o_url (str): Orchestrator url that links all valid projects
        filter_ex (bool, optional): If existing datasets should be filtered automatically. Defaults to False.
        filter_names (Optional[List[str]], optional): Names to be filtered from the dataset list. Defaults to None.
        query_url (str, optional): URL to the knowledgebase dataset query API.
        data_url (str, optional): URL to the knowledgebase dataset API.
        gitlab_url (str, optional): General Gitlab url. Defaults to "https://renkulab.io/gitlab".

    Returns:
        Tuple[List[str], List[str]]: 1. List with all matching non-existing dataset urls,
                                     2. List with all matching existig dataset urls.
    """
    all_ids, up_exist = get_field_by_dataset_property(
        string=keyword, url=query_url, filter_ex=filter_ex, filter_names=filter_names
    )
    up_exist = list(unique_everseen(up_exist))
    if len(all_ids) + len(up_exist) < 1:
        print(f"WARNING:No datasets found with keyword {keyword}")
        return [], []
    origin_infos = get_origin_dataset_ids(ids=all_ids, url=data_url)
    if len(origin_infos) + len(up_exist) < 1:
        print(
            "WARNING:Could not identify dataset sources.\n"
            f"Please check each of {data_url}{all_ids} to make sure they are the intended source"
        )
        return [], []
    if check_o_url:
        omni_ids = [
            check_orchestrator(
                data_info=info, o_url=o_url, gitlab_url=gitlab_url, n_latest=n_latest
            )
            for info in origin_infos
        ]
    else:
        omni_ids = [data_info["url"] for data_info in origin_infos]
    omni_ids = list(filter(None, omni_ids))
    if len(omni_ids) < 1:
        if len(up_exist) < 1:
            print(
                f"WARNING:No dataset with keyword {keyword} was associated to an omnibenchmark at ${o_url}\n"
                f"Please check if you specified the correct benchmark/url \n"
            )
        else:
            print(f"No new dataset with keyword {keyword} was found.")
    return omni_ids, up_exist  # type:ignore


def find_datasets_with_non_matching_keywords(
    keywords: List[str], include: Optional[List[str]] = None, remove: bool = True
):
    """Find datasets with non matching keywords

    Args:
        keywords (List[str]): Keywords to keep datasets with
        remove (bool, optional): Remove datasets and clean up. Defaults to True.
    """
    # datasets = Dataset.list()
    datasets = renku_api.renku_dataset_list()
    if include is not None:
        datasets = [dataset for dataset in datasets if dataset.name in include]
    data_filter = [
        dataset.name
        for dataset in datasets
        if not any(data_key in keywords for data_key in dataset.keywords)
    ]
    if remove:
        [renku_dataset_remove(data_filt) for data_filt in data_filter]


def update_datasets_by_keyword(
    keyword: str,
    o_url: str,
    filter_ex: bool = True,
    filter_names: Optional[List[str]] = None,
    query_url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    data_url: str = "https://renkulab.io/knowledge-graph/datasets/",
    gitlab_url: str = "https://renkulab.io/gitlab",
    check_o_url: bool = True,
    n_latest: int = 9,
    all: bool = True,
):
    """Import and/or update all datasets that match a certain keyword

    Args:
        keyword (str): Keyword to match datasets
        o_url (str):  Orchestrator url that links all valid projects
        filter_ex (bool, optional): If existing datasets should be filtered automatically. Defaults to True.
        filter_names (Optional[List[str]], optional): Names to be filtered from the dataset list. Defaults to None.
        query_url (_type_, optional): URL to the knowledgebase dataset query API.
        data_url (_type_, optional): URL to the knowledgebase dataset API.
        gitlab_url (_type_, optional): General Gitlab url. Defaults to "https://renkulab.io/gitlab".
        all (bool, optional): If all datasets with matching keyword should be imported.
    """
    imp_ids, up_names = get_data_url_by_keyword(
        keyword=keyword,
        o_url=o_url,
        filter_ex=filter_ex,
        filter_names=filter_names,
        query_url=query_url,
        data_url=data_url,
        gitlab_url=gitlab_url,
        check_o_url=check_o_url,
        n_latest=n_latest,
    )
    if all:
        for id in imp_ids:
            renku_dataset_import(uri=id)
    else:
        if len(imp_ids) > 0:
            renku_dataset_import(uri=imp_ids[0])
    for nam in up_names:
        print(f"Updated dataset {nam}.")
        renku_dataset_update(names=[nam])
        renku_save()
    find_datasets_with_non_matching_keywords(
        keywords=[keyword], include=up_names, remove=True
    )


def update_dataset_files(urls: List[str], dataset_name: str):
    """Add or update a list of files to/in a specified dataset.

    Args:
        urls (List[str]): File paths to add/update
        dataset_name (str): Dataset name to add files to/update files in
    """
    # datasets = Dataset.list()
    datasets = renku_api.renku_dataset_list()
    name_list = [dataset.name for dataset in datasets]
    if dataset_name not in name_list:
        print(
            f"WARNING: Dataset {dataset_name} does not exist.\n"
            f"Files will not be added to any dataset."
        )
        return
    dataset = [data for data in datasets if data.name == dataset_name][0]
    valid_urls = [url for url in urls if os.path.isfile(url)]
    dataset_files = [fi.path for fi in dataset.files]
    add_urls = [url for url in valid_urls if url not in dataset_files]
    if len(add_urls) > 0:
        renku_add_to_dataset(urls=add_urls, dataset_name=dataset_name)
        print(f"Added the following files to {dataset_name}:\n {add_urls}")
    return


def unlink_dataset_files(out_files: List[str], dataset_name: str, remove: bool = True):
    """Unlink files from a given dataset

    Args:
        out_files (List[str]): List of files to unlink
        dataset_name (str): Dataset name tounlink files from

    Raises:
        InputError: Dataset needs to refer to an existing dataset in the current project.
    """
    # datasets = Dataset.list()
    datasets = renku_api.renku_dataset_list()
    name_list = [dataset.name for dataset in datasets]
    if dataset_name not in name_list:
        raise InputError("Dataset {dataset_name} does not exist in this project.")
    dataset = [data for data in datasets if data.name == dataset_name][0]
    out_urls = [url for url in out_files if os.path.isfile(url)]
    dataset_files = [fi.path for fi in dataset.files]
    # What is with include as pattern? Can we replace all if renku doesn't complain for non-existing files?
    [
        renku_unlink_from_dataset(name=dataset_name, include=[out_url])
        for out_url in out_urls
        if out_url in dataset_files
    ]
    if remove:
        for out_url in out_urls:
            if os.path.exists(out_url):
                os.remove(out_url)



import gitlab
from gitlab.v4.objects.pipelines import ProjectPipeline
from gitlab.v4.objects.projects import Project
from omnibenchmark.management.data_commands import get_project_info_from_url
from typing import List, Optional
import base64
import yaml


def get_orchestrator_projects_from_cicd_yaml(
    o_url: str,
    gitlab_url: str = "https://renkulab.io/gitlab",
    target_branches: List[str] = ["master", "main"],
    exclude_stages: List[str] = ["build"],
) -> List[str]:
    """Get all projects associated to an orchestrator from the orchestrators gitlab cicd yaml file

    Args:
        o_url (str): Orchestrator url
        gitlab_url (_type_, optional): Gitlab url. Defaults to "https://renkulab.io/gitlab".
        target_branches (List[str], optional): branch(es) to use. Defaults to ["master", "main"].
        exclude_stages (List[str], optional): Stages to ignore. Defaults to ["build"].

    Returns:
        List[str]: Project urls associated to the
    """
    o_info = get_project_info_from_url(o_url)
    renku_git = gitlab.Gitlab(gitlab_url)
    o_git = renku_git.projects.get(o_info["identifier"])
    bs = o_git.branches.list()
    br = [b.name for b in bs if b.name in target_branches][0]
    ci_yml = o_git.files.get(file_path=".gitlab-ci.yml", ref=br)
    fc = base64.b64decode(ci_yml.content).decode("utf-8")
    fc = fc.replace("\\n", "\n")
    dct = yaml.safe_load(fc)
    stages = [stg for stg in dct["stages"] if stg not in exclude_stages]
    trigger_keys = ["stage", "trigger"]
    process = {
        run_nam: run_val
        for run_nam, run_val in dct.items()
        if isinstance(run_val, dict)
        and all(trigger_k in run_val.keys() for trigger_k in trigger_keys)
        and run_val["stage"] in stages
    }
    return [proc["trigger"]["project"] for proc in process.values()]


def get_project_infos(
    p_url: str, gitlab_url: str = "https://renkulab.io/gitlab"
) -> Optional[Project]:
    """Get project info from url

    Args:
        p_url (str): project url
        gitlab_url (_type_, optional): gitlab url. Defaults to "https://renkulab.io/gitlab".

    Returns:
        Project: An instance of gitlab API project class
    """
    p_info = get_project_info_from_url(p_url)
    if "message" in p_info.keys():
        print(f"Warning: Could not find project {p_url}")
        return None
    renku_git = gitlab.Gitlab(gitlab_url)
    return renku_git.projects.get(p_info["identifier"])


def get_last_pipe(
    project: Project, exclude_push: bool = True
) -> Optional[ProjectPipeline]:
    """Get the last pipeline from an project instance

    Args:
        project (Project): An instance of the gitlab api Project class
        exclude_push (bool, optional): if pipelines generated by push shall be excluded. Defaults to True.

    Returns:
        Optional[ProjectPipeline]: AN instance of the gitlab api projectpipeline class
    """
    pipes = project.pipelines.list(order_by="updated_at", get_all=True)
    if exclude_push:
        pipes = [p for p in pipes if p.source != "push"]
    if len(pipes) > 0:
        return pipes[0]  # type:ignore
    else:
        print(f"Warning: Did not find any valid pipeline for project {project.web_url}")
        return None



from __future__ import annotations

# from typing_extensions import TypedDict, Required
from typing import Callable, Mapping, TypedDict
from omnibenchmark.utils.exceptions import InputError
import yaml
from collections import defaultdict
from omnibenchmark.core.omni_object import OmniObject
from omnibenchmark.core.output_classes import OmniCommand, OmniOutput
from omnibenchmark.core.input_classes import OmniInput, OmniParameter
from os import PathLike
from typing import Any, Optional, List, Union
from omnibenchmark.utils.decorators import option_dict_none, option_dict_list
from omnibenchmark.utils.user_input_checks import flatten, empty_object_to_none
import logging
from typeguard import check_type

logger = logging.getLogger("omnibenchmark.build_object")


# Config definitions:


class ConfigData(TypedDict, total=False):
    name: str
    title: Optional[str]
    description: Optional[str]
    keywords: Optional[Union[str, List[str]]]


class ConfigInput(TypedDict, total=False):
    keywords: Optional[Union[str, List[str]]]
    files: Optional[Union[str, List[str]]]
    prefix: Optional[dict]
    input_files: Optional[dict]
    default: Optional[str]
    filter_names: Optional[Union[str, List[str]]]


class ConfigOutput(TypedDict, total=False):
    template: Optional[str]
    files: Optional[Mapping]
    file_mapping: Optional[Mapping]
    default: Optional[Mapping]
    filter_json: Optional[str]
    sort_keys: Optional[str]
    template_fun: Optional[Callable]
    template_vars: Optional[Mapping]


class ConfigParam(TypedDict, total=False):
    # names: Required[List[str]]
    names: List[str]
    keywords: Optional[Union[str, List[str]]]
    default: Optional[Mapping]
    values: Optional[Mapping]
    combinations: Optional[Mapping]
    filter: Optional[Mapping]


class ConfigDict(TypedDict, total=False):
    # data: Required[ConfigData]
    data: ConfigData
    script: Optional[str]
    interpreter: Optional[str]
    command_line: Optional[str]
    benchmark_name: Optional[str]
    orchestrator: Optional[str]
    inputs: Optional[ConfigInput]
    outputs: Optional[ConfigOutput]
    parameter: Optional[ConfigParam]


# Functions to build OmniObjects from yaml/config_dict


@option_dict_none
def get_keys(map: Mapping[str, Any]) -> List[str]:
    return list(map.keys())


@option_dict_list
def get_items(map: Mapping[str, Any]) -> List[tuple]:
    return list(map.items())


@option_dict_list
def get_values(map: Mapping[str, Any]) -> List[tuple]:
    return list(map.values())


def into_list(obj: Union[str, list]) -> list:
    if not isinstance(obj, list):
        obj = [obj]
    return obj


@option_dict_none
def build_omni_input_from_config_inputs(
    config_inputs: ConfigInput,
) -> Optional[OmniInput]:
    """Build an OmniInput object from a ConfigInput dictionary

    Args:
        config_inputs (ConfigInput): A ConfigInput dictionary

    Returns:
        Optional[OmniInput]: An OmniInput object
    """
    inputs: defaultdict = defaultdict(list, config_inputs)
    in_pre = inputs["prefix"]
    in_names = empty_object_to_none(into_list(inputs["files"]))
    prefix = empty_object_to_none(
        {name: into_list(pre) for name, pre in get_items(in_pre)}
    )
    keywords = into_list(inputs["keywords"])
    in_files = empty_object_to_none(inputs["input_files"])
    in_file_types = empty_object_to_none(
        list(set(flatten([list(in1.keys()) for in1 in get_values(in_files)])))
    )
    default_in = empty_object_to_none(inputs["default"])
    filter_names = empty_object_to_none(into_list(inputs["filter_names"]))
    in_names = in_names or get_keys(prefix) or in_file_types

    if in_names is not None:
        return OmniInput(
            names=in_names,
            prefix=prefix,
            keyword=keywords,
            input_files=in_files,
            default=default_in,
            filter_names=filter_names,
        )
    else:
        logger.warning(
            "Warning: Skipped inputs because of missing information. Please specify at least names."
        )
        return None


@option_dict_none
def build_omni_parameter_from_config_params(
    config_params: ConfigParam,
) -> Optional[OmniParameter]:
    """Build an OmniParameter object from a ConfigParam dictionary

    Args:
        config_params (ConfigParam): A ConfigParam dictionary

    Returns:
        Optional[OmniParameter]: An OmniParameter object
    """
    params: defaultdict = defaultdict(list, config_params)
    param_names = empty_object_to_none(into_list(params["names"]))
    param_values = empty_object_to_none(params["values"])
    param_combs = empty_object_to_none(
        [vals for vals in get_values(params["combinations"])]
    )
    param_comb_types = empty_object_to_none(
        list(
            set(
                flatten(
                    [list(in1.keys()) for in1 in get_values(params["combinations"])]
                )
            )
        )
    )
    default_params = empty_object_to_none(params["default"])
    param_names = (
        param_names
        or get_keys(param_values)
        or get_keys(default_params)
        or param_comb_types
    )
    keywords = into_list(params["keywords"])
    param_filter = empty_object_to_none(params["filter"])
    if param_names is not None:
        return OmniParameter(
            names=param_names,
            filter=param_filter,
            keyword=keywords,
            values=param_values,
            combinations=param_combs,
            default=default_params,
        )
    else:
        logger.warning(
            "Warning: Skipped parameter because of missing information. Please specify at least names."
        )
        return None


def build_omni_output_from_config(
    config: ConfigDict,
    omni_input: Optional[OmniInput] = None,
    omni_parameter: Optional[OmniParameter] = None,
) -> Optional[OmniOutput]:
    """Build an OmniOutput from a ConfigDict

    Args:
        config (ConfigDict): A ConfigDict
        omni_input (Optional[OmniInput], optional): An OmniInput object. Defaults to None.
        omni_parameter (Optional[OmniParameter], optional): An Omniparameter object. Defaults to None.

    Returns:
        Optional[OmniOutput]: An OmniOutput object.
    """
    config_in: defaultdict = defaultdict(dict, config)
    config_outputs = defaultdict(list, config_in["outputs"])
    out_files = config_outputs["files"]
    out_names = get_keys(out_files)
    out_ends = empty_object_to_none(
        {out_names: vals["end"] for out_names, vals in get_items(out_files)}
    )
    template = config_outputs["template"]
    if len(template) == 0:
        template = "data/${name}/${name}_${unique_values}_${out_name}.${out_end}"
    try:
        name = config_in["data"]["name"]
    except Exception:
        raise (
            InputError(
                f"Invalid config file: {config}\n"
                f"Please provide a config.yaml file with a value for data: name:"
            )
        )
    file_map = config_outputs["file_mapping"]
    file_mapping = empty_object_to_none([vals for vals in get_values(file_map)])
    default_out = empty_object_to_none(config_outputs["default"])
    filter_json = empty_object_to_none(config_outputs["filter_json"])
    sort_keys = True if (len(config_outputs["sort_keys"]) == 0) else eval(config_outputs["sort_keys"])
    template_fun = config_outputs["template_fun"]
    template_fun = None if isinstance(template_fun, List) else template_fun
    template_vars = empty_object_to_none(config_outputs["template_vars"])

    if not isinstance(out_names, type(None)):
        return OmniOutput(
            name=name,
            out_names=out_names,
            output_end=out_ends,
            out_template=template,
            file_mapping=file_mapping,  # type:ignore
            inputs=omni_input,
            parameter=omni_parameter,
            default=default_out,  # type:ignore
            filter_json=filter_json,  # type:ignore
            sort_keys=sort_keys,
            template_fun=template_fun,
            template_vars=template_vars,  # type:ignore
        )
    else:
        logger.warning(
            "Warning: Skipped outputs because of missing information. Please specify at least output names."
        )
        return None


def build_omni_command_from_config(
    config: ConfigDict, omni_output: Optional[OmniOutput] = None
) -> Optional[OmniCommand]:
    """Build an OmniCommand object from a ConfigDict_

    Args:
        config (ConfigDict): A configDict instance
        omni_output (Optional[OmniOutput], optional): An OmniOutput object. Defaults to None.

    Returns:
        Optional[OmniCommand]: An OmniCommand object
    """
    if omni_output is None:
        omni_output = build_omni_output_from_config(config)
    config_in: defaultdict = defaultdict(list, config)
    obj_script = empty_object_to_none(config_in["script"])
    obj_interpreter = empty_object_to_none(config_in["interpreter"])
    obj_command_line = empty_object_to_none(config_in["command_line"])
    if not isinstance(obj_script, type(None)):
        return OmniCommand(
            script=obj_script,
            interpreter=obj_interpreter,
            command_line=obj_command_line,
            outputs=omni_output,
        )
    else:
        logger.warning(
            "Warning: Skipped command because of missing information. Please specify at least script: script path."
        )
        return None


def build_omni_object_from_config(config: ConfigDict) -> OmniObject:
    """Build an OmniObject from a Config dict instance

    Args:
        config (ConfigDict): A ConfigDict instance.

    Returns:
        OmniObject: An OmniObject object.
    """
    config_in: defaultdict = defaultdict(dict, config)
    omni_input = build_omni_input_from_config_inputs(config_in["inputs"])
    omni_parameter = build_omni_parameter_from_config_params(config_in["parameter"])
    omni_output = build_omni_output_from_config(
        config, omni_input=omni_input, omni_parameter=omni_parameter
    )
    omni_command = build_omni_command_from_config(
        config=config, omni_output=omni_output
    )
    config_data = defaultdict(list, config_in["data"])
    obj_name = empty_object_to_none(config_data["name"])
    obj_key = empty_object_to_none(into_list(config_data["keywords"]))
    obj_title = empty_object_to_none(config_data["title"])
    obj_description = empty_object_to_none(config_data["description"])
    obj_script = empty_object_to_none(config_in["script"])
    obj_bench_name = empty_object_to_none(config_in["benchmark_name"])
    obj_orchestrator = empty_object_to_none(config_in["orchestrator"])

    omni_object = OmniObject(
        name=obj_name,  # type:ignore
        keyword=obj_key,
        title=obj_title,  # type:ignore
        description=obj_description,  # type:ignore
        script=obj_script,
        benchmark_name=obj_bench_name,
        orchestrator=obj_orchestrator,
        inputs=omni_input,
        command=omni_command,
        outputs=omni_output,
        parameter=omni_parameter,
    )
    return omni_object


def get_omni_object_from_yaml(yaml_file: PathLike) -> OmniObject:
    """Build an OmniObject from defenitions of a yaml file

    Args:
        yaml_file (PathLike): Path to yaml file

    Returns:
        OmniObject: An OmniObject object.
    """
    with open(yaml_file) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    check_type("config", config, ConfigDict)
    return build_omni_object_from_config(config)



from omnibenchmark.utils.exceptions import InterpreterError
from typing import Mapping, Union, List, Optional
import os.path
from os import PathLike


def get_interpreter_from_extension(ext: str) -> str:
    """Automatically gets the interpreter to run a script with from the scripts file extension.

    Args:
        ext (str): File extension from the script, that should be run as workflow

    Raises:
        InterpreterError: Error, if no known interpreter can be identified.

    Returns:
        str: Interpreter type, e.g. R, python, bash, julia
    """

    interpreter = {"r": "R", "py": "python", "sh": "bash", "jl": "julia"}.get(
        ext.lstrip(".").lower(), None
    )

    if interpreter is None:
        raise InterpreterError(
            "Interpreter could not be identified from extention {}. Please specify command explicitly.".format(
                ext
            )
        )
    else:
        return interpreter


def get_command_start_from_interpreter(interpreter: str) -> str:
    """Automatically generates the command start to run a script with from the interpreter name.

    Args:
        interpreter (str): Interpreter to run a script with, e.g., R, python

    Raises:
        InterpreterError: Error, if no known interpreter can be identified.

    Returns:
        str: Command start to run the script with
    """

    command_start = {
        "r": "Rscript --vanilla",
        "python": "python",
        "bash": "bash",
        "julia": "julia",
    }.get(interpreter.lower(), None)

    if command_start is None:
        raise InterpreterError(
            f"Command start could not be identified from interpreter {interpreter}."
            "Please specify interpreter and command explicitly."
        )
    else:
        return command_start


def parse_command_line_args(arg_map: Mapping[str, str]) -> List[str]:
    """Parse arguments from a mapping as command line arguments of a renku run.

    Args:
        arg_map (Mapping[str, str]): Mapping of arguments

    Returns:
        List[str]: List of parsed commands
    """
    return [
        '--{} "{}"'.format(str(arg_nam), arg_val)
        for arg_nam, arg_val in arg_map.items()
    ]


def automatic_command_generation(
    script: Union[PathLike, str],
    interpreter: Optional[str] = None,
    inputs: Optional[Mapping[str, str]] = None,
    parameters: Optional[Mapping[str, str]] = None,
    outputs: Optional[Mapping[str, str]] = None,
) -> str:
    """Automatic generate a command line command to run with renku run

    Args:
        script (PathLike): Path to script to run
        interpreter (str, optional): Interpreter to use to run the script. Defaults to None.
        inputs (Mapping, optional): Default inputs to run the script. Defaults to None.
        parameters (Mapping, optional): Default parameter to run the script with. Defaults to None.
        outputs (Mapping, optional): Default outputs generated. Defaults to None.

    Returns:
        str: A command line command to run the script with and parse all inputs/parameter/output names
             as commandline options (format --name value)
    """
    if interpreter is None:
        ext = os.path.splitext(script)[1]
        interpreter = get_interpreter_from_extension(ext)

    command_start = get_command_start_from_interpreter(interpreter)

    arg_list: List[str] = []
    arg_list.extend(
        parse_command_line_args(arg_map)  # type: ignore
        for arg_map in [inputs, outputs, parameters]
        if arg_map is not None
    )
    flat_list = [item for sublist in arg_list for item in sublist]
    command_list = [command_start, script]
    command_list.extend(flat_list)
    command_line = " ".join(command_list)  # type:ignore
    return command_line



"""Functions to facilitate automatic input generation from file/object, usually config.yaml"""

from typing import Dict, Mapping, List, Optional
from omnibenchmark.utils.exceptions import ParameterError
from collections import defaultdict
from renku.ui.api.models.dataset import Dataset
from omnibenchmark.renku_commands import renku_api
from difflib import SequenceMatcher
import hashlib
import re
import os
import json


def find_stem(arr):
    """Find a common substring from a list of strings

    Args:
        arr (Array): Array of strings

    Returns:
        str: longest common substring
    """
    n = len(arr)
    # Take first word from array
    # as reference
    s = arr[0]
    len_s = len(s)
    res = ""
    for i in range(len_s):
        for j in range(i + 1, len_s + 1):

            # generating all possible substrings
            # of our reference string arr[0] i.e s
            stem = s[i:j]
            k = 1
            for k in range(1, n):

                # Check if the generated stem is
                # common to all words
                if stem not in arr[k]:
                    break

                # If current substring is present in
                # all strings and its length is greater
                # than current result
                if k + 1 == n and len(res) < len(stem):
                    res = stem

    return res


def best_match_name_seq(map_dict: Mapping[str, str]) -> str:
    """Get the longest common substring from a mapping of file names without directories

    Args:
        map_dict (Mapping[str, str]): Mapping of file names

    Returns:
        str: Longest common substring
    """
    na_list = [os.path.basename(fi_nam) for fi_nam in map_dict.values()]
    nam_list = [os.path.splitext(nam)[0] for nam in na_list]
    name_list = [os.path.splitext(nam)[0] for nam in nam_list]
    com_sub = find_stem(name_list)
    return com_sub.rstrip("_- .")


def get_name_hash_from_input_dict(infile_dict: Mapping[str, str]) -> str:
    """Get  the md5 hash of the joined values of a dictionary (e.g. of all input file names in an input file dict)

    Args:
        infile_dict (Mapping[str,str]): A dictionary with file types as key and file names as values.

    Returns:
        (str): The md5 hash of the concatenated file names
    """
    in_string = "".join(sorted(list(infile_dict.values())))
    hash_object = hashlib.md5(in_string.encode())
    return hash_object.hexdigest()


def match_files_by_name(
    file_type_dict: Mapping[str, List[str]]
) -> Mapping[str, Mapping]:
    """Find corresponding files by best matching names

    Args:
        file_type_dict (Mapping[str, List[str]]): Dictionary specifying all file types and their corresponding files

    Returns:
        Mapping[str, Mapping]: Mapping of best matches for all files between file types.
    """
    match_dict: Dict = {}
    fi_types = list(file_type_dict.keys())
    fi_start = max(file_type_dict, key=lambda x: len(set(file_type_dict[x])))
    fil_types = [fi_type for fi_type in fi_types if not fi_type == fi_start]
    for fi_idx, fi in enumerate(file_type_dict[fi_start]):
        group_nam = "inst" + str(fi_idx)
        match_dict[group_nam] = {fi_start: fi}
        for fi_type in fil_types:
            fi_path_list = (
                [file_type_dict[fi_type]]
                if not isinstance(file_type_dict[fi_type], list)
                else file_type_dict[fi_type]
            )
            fi_top = ""
            seq_top = 0
            for fi_path in fi_path_list:  # type:ignore
                fi_name = os.path.basename(fi_path)
                seq_match = (
                    SequenceMatcher(None, os.path.basename(fi), fi_name)  # type:ignore
                    .find_longest_match()
                    .size
                )
                if seq_match > seq_top:
                    seq_top = seq_match
                    top_blocks = SequenceMatcher(
                        None, os.path.basename(fi), fi_name
                    ).get_matching_blocks()
                    fi_top = fi_path
                elif seq_match == seq_top:
                    seq_blocks = SequenceMatcher(
                        None, os.path.basename(fi), fi_name
                    ).get_matching_blocks()
                    for seq, top in zip(seq_blocks, top_blocks):
                        if seq.size > top.size:
                            fi_top = fi_path
                            top_blocks = seq_blocks
                            break
                        if top.size > seq.size:
                            break
            match_dict[group_nam][fi_type] = fi_top
    com_dict = {
        get_name_hash_from_input_dict(match_dict[m_key])[0:5]
        + "_"
        + best_match_name_seq(match_dict[m_key]): match_dict[m_key]
        for m_key in match_dict.keys()
    }
    return com_dict


def check_dataset_name(input_dict: Mapping[str, str], data_name: str) -> str:
    """Check if data_name includes the longest matching sequence of all input files

    Args:
        input_dict (Mapping[str, str]): An input file type- name mapping
        data_name (str): Dataset name to check

    Returns:
        str: Either the dataset name, if it is equal the longest matching sequence
             or a name of the dataset name, longest match seq and an input file name hash.
    """
    best_match = best_match_name_seq(input_dict)
    if best_match == data_name:
        return data_name
    else:
        new_name = (
            data_name
            + "_"
            + get_name_hash_from_input_dict(input_dict)[0:5]
            + "_"
            + best_match
        )
        return new_name.rstrip("_- .")


def remove_duplicated_inputs(pat_list: List, in_files: List[str]) -> List[str]:
    """Removes elements from input files if their names are identical except for the file prefix pattern
       when these pattern are associated with the same input file type. Will automatically select the first match.

    Args:
        pat_list (List): List of file prefix pattern generated by re.compile
        in_files (List[str]): All input files that matched one of the pattern in pat_list

    Returns:
        List[str]: input file list with duplicated files removed.
    """
    simp_infiles = in_files
    for pat in pat_list:
        simp_infiles = [re.sub(pat, "", s) for s in simp_infiles]
    stem_ind = defaultdict(list)
    for i, stem in enumerate(simp_infiles):
        stem_ind[stem].append(i)
    ind_keep = [ind[0] for ind in stem_ind.values()]
    return [in_files[ind] for ind in ind_keep]


def get_input_files_from_prefix(
    input_prefix: Mapping[str, List[str]],
    keyword: List[str],
    filter_names: Optional[List[str]] = None,
    filter_duplicated: bool = True,
) -> Mapping[str, Mapping]:
    """Find input files by prefix

    Args:
        input_prefix (Mapping[str, List[str]]): Dictionary with all file types and their valid prefixes.
        keyword (List[str]): Keyword of datasets that contain input files
        filter_names (Optional[List[str]]): Dataset names to be ignored when querying input files
        filter_duplicated (bool): If duplicated inputs from the same dataset and prefix pattern
                                  associated to the same input file types shall automatically be removed.

    Returns:
        Mapping[str, Mapping]: Input file types with their corresponding files.
    """
    input_files: Dict = {}
    # datasets = Dataset.list()
    datasets = renku_api.renku_dataset_list()
    key_data = [
        dataset
        for dataset in datasets
        if any(key in keyword for key in dataset.keywords)
    ]
    for data in key_data:
        tmp_dict: Dict = {}
        input_files[data.name] = {}
        for file_type, prefixes in input_prefix.items():
            prefix_list = (
                [prefixes]  # type: ignore
                if not isinstance(prefixes, list)
                else prefixes
            )
            pat_list = [re.compile(pattern) for pattern in prefix_list]
            in_file = [
                fi.path
                for fi in data.files
                if any(
                    pattern.search(os.path.basename(fi.path)) for pattern in pat_list
                )
            ]
            if len(in_file) > 1:
                if len(pat_list) > 1 and filter_duplicated:
                    in_file = remove_duplicated_inputs(
                        pat_list=pat_list, in_files=in_file
                    )
                tmp_dict[file_type] = in_file
            elif len(in_file) < 1:
                print(
                    f"WARNING:Could not find any input file matching the following pattern {prefixes}.\n"
                    f"Please make sure you specified a correct prefix! \n"
                    f"Input dataset {data.name} will be ignored!"
                )
                break
            else:
                input_files[data.name][file_type] = in_file[0]
        if len(tmp_dict) > 0:
            tmp_dict.update(input_files[data.name])
            group_dict = match_files_by_name(tmp_dict)
            n = len(in_file) if len(in_file) <= 5 else 5
            print(
                f"WARNING: Ambigous input files. Found {in_file[0:n]}, ...\n"
                f"Automatic group detection for files in {data.name}.\n"
                f"Please check matches to ensure correct groups: 'omni_obj.inputs.input_files'"
            )
            del input_files[data.name]
            group_data_dict = {
                data.name + "_" + k: group_dict[k] for k in group_dict.keys()
            }
            input_files.update(group_data_dict)
    file_types = input_prefix.keys()

    # Filter incomplete, non existing and explicitly filtered data
    incomplete_data = [
        data
        for data in input_files.keys()
        if not all(fi_type in input_files[data].keys() for fi_type in file_types)
    ]
    non_exist_data = [
        data
        for data in input_files.keys()
        if not all(os.path.exists(fi) for fi in input_files[data].values())
    ]
    rm_data = incomplete_data + non_exist_data
    if filter_names is not None:
        filter_list = [
            filter_nam
            for filter_nam in filter_names
            if filter_nam in input_files.keys()
        ]
        rm_data = rm_data + filter_list
    for data in rm_data:
        del input_files[data]
    dataset_names = [data.name for data in key_data]

    for input_nam in list(input_files.keys()):
        if input_nam in dataset_names:
            new_name = check_dataset_name(input_files[input_nam], input_nam)
            if new_name != input_nam:
                input_files[new_name] = input_files[input_nam]
                del input_files[input_nam]

    return input_files


def get_parameter_from_dataset(
    names: List[str], keyword: List[str]
) -> Mapping[str, List]:
    """Get parameter from a renku dataset

    Args:
        names (List[str]): Parameter names
        keyword (List[str]): Keyword of the parameter dataset.

    Raises:
        ParameterError: _description_

    Returns:
        Mapping[str, List]: Mapping with parameters and all possible values
    """
    values: Dict = {}
    datasets = renku_api.renku_dataset_list()
    # datasets = Dataset.list()
    key_data = [
        dataset
        for dataset in datasets
        if any(key in keyword for key in dataset.keywords)
    ]
    for data in key_data:
        if not len(data.files) == 1:
            raise ParameterError(
                f"Could not identify parameter json file."
                f"Please check the specified keyword {keyword} and dataset {data}."
            )
        param_file = data.files[0]
        with open(param_file.path) as f:
            param_json = json.load(f)

        for nam in names:
            if nam not in param_json.keys():
                print(
                    f"WARNING: Could not find a matching parameter for {nam}.\n"
                    f"The following parameter exist in {data}: {param_json.keys()} \n"
                    f"Please post an issue to request missing parameter."
                )
                continue
            values[nam] = param_json[nam]

    print(f"The following parameter are used: {values.keys()}")
    return values


def drop_none_elements(l: List):
    return [element for element in l if element is not None]



""" All kind of check or test functions to ensure correct variable inputs beyont type hinting"""

from typing import List, Iterable, Mapping, Union, Optional, overload
from omnibenchmark.utils.exceptions import InputError


def flatten(nested_list: List[Iterable]) -> List:
    """Flatten a nested list into a single list

    Args:
        nested_list (List[List]): A nested list

    Returns:
        List: A flat list
    """
    return [item for sublist in nested_list for item in sublist]


def mixed_flatten(mixed_list: List[Iterable]) -> List:
    """Flatten a list of list elements and none list elements

    Args:
        mixed_list (List[Iterable]): A list with mixed and none mixed elements

    Returns:
        List: A flat list
    """
    nested = [
        item for sublist in mixed_list if type(sublist) is list for item in sublist
    ]
    unnested = [item for item in mixed_list if type(item) is not list]
    return nested + unnested


def check_name_matching(entry1: Iterable[str], entry2: Iterable[str]):
    """Check if all entries of a list (entry2) are part of a second list (entry1) of entries

    Args:
        entry1 (List[str]): List to with variables to check
        entry2 (List[str]): List containing the variables that are allowed

    """
    if any(val not in entry1 for val in entry2):
        uniq_entry2 = list(set(entry2))
        raise InputError(f"Expected inputs from ${entry1}, but found ${uniq_entry2}")
    else:
        pass


@overload
def empty_object_to_none(obj: List) -> Optional[List]:
    ...


@overload
def empty_object_to_none(obj: Mapping) -> Optional[Mapping]:
    ...


def empty_object_to_none(obj: Union[List, Mapping]) -> Optional[Union[List, Mapping]]:
    if len(obj) == 0:
        return None
    else:
        return obj


def parse_explicit_inputs(obj: Optional[Mapping]) -> List:
    if obj is None:
        return []
    else:
        return list(obj.values())


def check_default_parameter(
    default_params: Optional[Mapping], param_combinations: Optional[List[Mapping]]
) -> Optional[Mapping]:
    """Check if the specified default parameter are part of the available parameter combinations

    Args:
        default_params (Optional[Mapping]): Default parameter mapping
        param_combinations (Optional[List[Mapping]]): List of all available parameter combinations

    Raises:
        InputError: If parameter defaults are specified, but no parameter dataset is detected.

    Returns:
        Optional[Mapping]: Default parameter mapping
    """
    if default_params is not None and (
        param_combinations is None or len(param_combinations) < 1
    ):
        raise InputError(
            "No parameter input detected, but default specified.\n"
            "Did you intend to run different parameter? If yes please check the parameter input.\n"
            "Otherwise please do not specify default parameter outside your script."
        )
    if (
        default_params is None
        or default_params in param_combinations  # type: ignore
        or len(default_params) == 0
    ):
        return default_params
    else:
        print(
            f"WARNING: The specified parameter defaults are not part of the parameter space.\n"
            f"The specified defaults will be ignored!\n"
            f"Please check your defaults:\n {default_params}\n and the parameter space:\n {param_combinations}"
        )
        return param_combinations[0]  # type: ignore


def rm_none_from_list(in_list: List) -> List:
    return [element for element in in_list if element is not None]


def elements_to_string(d: Mapping) -> Mapping:
    """Convert all non iterable elements in a dict to strings

    Args:
        d (dict): Dictionary to convert

    Returns:
        Mapping: Converted dictionary
    """
    for k, v in d.items():
        if not isinstance(v, Iterable):
            d[k] = str(v)  # type:ignore
        if type(v) is list:
            d[k] = [e if isinstance(e, Iterable) else str(e) for e in v]  # type:ignore
    return d


def elements_to_list(d: Mapping) -> Mapping:
    """Convert all non list elements in a dict to lists

    Args:
        d (dict): Dictionary to convert

    Returns:
        Mapping: Converted dictionary
    """
    for k, v in d.items():
        if not isinstance(v, list):
            d[k] = [v]  # type:ignore
    return d



"""Commands related to execute a script with renku workflow/run commands"""

from typing import List, Optional, Union
from omnibenchmark.core.input_classes import OutMapping
from omnibenchmark.utils.auto_run import (
    get_file_mapping_from_out_files,
    get_file_type_dict,
    get_file_name_dict,
)
from renku.domain_model.workflow.plan import Plan as PlanDomainModel
from omnibenchmark.utils.exceptions import InputError, ParameterError
from omnibenchmark.management import wflow_checks as wflow
from omnibenchmark.management.data_commands import unlink_dataset_files
import omnibenchmark.renku_commands.workflows as omni_wflow
from omnibenchmark.utils.auto_output import get_default, convert_values_to_string
from omnibenchmark.utils.user_input_checks import (
    parse_explicit_inputs,
    empty_object_to_none,
    flatten,
)
from omnibenchmark.core.output_classes import OmniCommand, OmniOutput, OmniPlan
from omnibenchmark.renku_commands.general import renku_save
from renku.command.view_model.plan import PlanViewModel
from renku.api import Activity, Plan
from renku.domain_model.project_context import project_context
import os
from os import PathLike
import logging

logger = logging.getLogger("omnibenchmark.management.run_commands")


def update_activity(out_map: OutMapping):
    """update existing activities

    Args:
        out_map (OutMapping): Output Mapping with the path to update activities for
    """
    if out_map["output_files"] is not None:
        out_paths = list(out_map["output_files"].values())
    else:
        out_paths = None
    omni_wflow.renku_update_activity(paths=out_paths)


def create_activity(out_map: OutMapping, omni_plan: OmniPlan):
    """Run an activity to generate a specific output from a plan

    Args:
        out_map (OutMapping): Output mapping with inputs, parameter and output files
        omni_plan (OmniPlan): plan, workflow description to use

    Raises:
        NameError: All elements from outmapping need to have an unique mapping in the plan
    """
    plan = omni_plan.plan
    map_dict = omni_plan.param_mapping
    file_dict = get_file_type_dict(out_map)
    if map_dict is None or any(
        file_type not in list(map_dict.keys()) for file_type in list(file_dict.keys())
    ):
        raise NameError(
            f"Could not assign all workflow items from: ${omni_plan.param_mapping}."
            f'Please check "renku workflow show ${plan.name}".'
            f"Update OmniObject.omni_plan.param_mapping accordingly."
        )
    params = [
        map_dict[str(file_type)] + "=" + str(file_path)
        for file_type, file_path in file_dict.items()
    ]
    workflow = omni_wflow.renku_workflow_execute(name_or_id=plan.id, set_params=params)
    return workflow


def manage_renku_plan(
    omni_plan: Optional[OmniPlan],
    command: OmniCommand,
    output: Optional[OmniOutput],
    name: Optional[str] = None,
    description: Optional[str] = None,
    keyword: Optional[str] = None,
) -> OmniPlan:
    """Check if a renku plan for a set of output files exist and generate one from default parameter if not.

    Args:
        output (Optional[OmniOutput]): An OmniOutput instance with all output files and their mappimg specified.
        command (OmniCommand): OmniCommand object with command line and outputs specified.
        name (str, optional): Workflow name. Defaults to None.
        description (str, optional): Workflow description. Defaults to None.
        keyword (str, optional): Workflow keyword. Defaults to None.

    Returns:
        OmniPlan: A renku planview model with param_mapping
    """
    if output is None:
        output = command.outputs
    if output is None:
        raise InputError(
            "Please specify outputs to be generated.\n"
            "You can either specify command.outputs or outputs directly"
        )
    out_files = get_all_output_file_names(output)
    plan = wflow.check_plan_exist(out_files)
    if plan is not None:
        plan_view = PlanViewModel.from_plan(plan)

    if plan is None:
        if command.outputs is not output:
            command.outputs = output
            command.update_command()
        if command.command_line is None:
            raise InputError(
                "Can not generate a workflow without command and outputs specified.\n"
                "Please make sure to provide an object of class OmniCommand with outputs specified."
            )
        def_input = parse_explicit_inputs(
            empty_object_to_none(convert_values_to_string(command.input_val))
        )
        def_parameter = parse_explicit_inputs(
            empty_object_to_none(convert_values_to_string(command.parameter_val))
        )
        if len(set(def_parameter)) < len(def_parameter):
            def_parameter = []
        def_output = parse_explicit_inputs(
            empty_object_to_none(get_default(command.outputs))
        )
        result = omni_wflow.renku_workflow_run(
            command_line=command.command_line,
            name=name,
            description=description,
            keyword=keyword,
            explicit_inputs=def_input,
            explicit_parameters=def_parameter,
            explicit_outputs=def_output,
        )

        plan_view = result.output

    if omni_plan is None or not plan_view == omni_plan.plan:
        omni_plan = OmniPlan(plan=plan_view)
        plan_outs = [plan_out.default_value for plan_out in plan_view.outputs]
        file_mapping_plan = get_file_mapping_from_out_files(
            out_files=plan_outs, file_mapping=output.file_mapping
        )[0]
        file_list = [
            file_map
            for file_map in [
                file_mapping_plan["output_files"],
                file_mapping_plan["input_files"],
                file_mapping_plan["parameter"],
            ]
            if file_map is not None
        ]
        file_dict = get_file_name_dict(file_mapping=file_list)
        omni_plan.param_mapping = omni_plan.predict_mapping_from_file_dict(
            file_dict=file_dict
        )

    return omni_plan


def check_omni_command(
    command: Optional[OmniCommand],
    script: Optional[Union[PathLike, str]],
    outputs: Optional[OmniOutput],
) -> OmniCommand:
    """Generate and check a command from a specified script and outputs

    Args:
        command (Optional[OmniCommand]): An Omnicommand object to check
        script (Optional[Union[PathLike, str]]): A script to generate the command for.
        outputs (Optional[OmniOutput]): OmniOutput that is associated to the command.

    Raises:
        InputError: Minimal a script to run the command on needs to be specified.

    Returns:
        OmniCommand: An OmniCommand object
    """
    if command is None:
        if script is None:
            raise InputError(
                "Can not run a command without a specified script. Please specify what script to run."
            )
        command = OmniCommand(script=script, outputs=outputs)
    return command


def get_all_output_file_names(output: OmniOutput) -> List[str]:
    """Get a list of all files specified as outputs

    Args:
        output (OmniOutput): An OmniOutputr object

    Returns:
        List[str]: A list of output file names
    """
    if output.file_mapping is not None:
        return flatten(
            [
                list(fi_mapping["output_files"].values())
                for fi_mapping in output.file_mapping
                if fi_mapping["output_files"] is not None
            ]
        )
    else:
        return []


def manage_renku_activities(
    outputs: OmniOutput, omni_plan: OmniPlan, save: bool = False
):
    """Manage renku activities by updating existing ones and generating new activities for output files without.

    Args:
        outputs (OmniOutput): An OmniOutput object
        omni_plan (OmniPlan): A plan/workflow description with mappings to the output
    """
    project_context.clear()
    out_files = get_all_output_file_names(outputs)
    activity_out = wflow.filter_activity_exist(out_files)
    activity_map = get_file_mapping_from_out_files(
        out_files=activity_out, file_mapping=outputs.file_mapping
    )
    if outputs.file_mapping is None:
        outputs.file_mapping = []

    no_activities = [out for out in outputs.file_mapping if out not in activity_map]

    for out_map in no_activities:
        create_activity(out_map, omni_plan)
        if save:
            renku_save(message="new activity, no image rebuild")

    for activity in activity_map:
        update_activity(activity)
        if save:
            renku_save(message="update activity, no image rebuild")


def check_output_directories(out_files: List[str]):
    """Check if output directoroes exist and generate them if not

    Args:
        out_files (List[str]): List of output files
    """
    out_dir_list = [os.path.dirname(out_file) for out_file in out_files]
    out_dirs = set(out_dir_list)
    for dir in out_dirs:
        os.makedirs(dir, exist_ok=True)


def revert_run(
    out_files: Optional[List[str]] = None,
    in_files: Optional[List[str]] = None,
    plan_id: Optional[str] = None,
    dataset_name: Optional[str] = None,
    remove: bool = True,
):
    """Remove a plan and revert all associated activities. Basically revert all actions of a specific omni_obj.run_renku().

    Args:
        out_files (Optional[List[str]], optional): Output files associated to the activities to revert. Defaults to None.
        in_files (Optional[List[str]], optional): Input files associated to the activities to revert. Defaults to None.
        plan_id (Optional[str], optional): Plan to revert including all associated activities. Defaults to None.
        dataset_name (Optional[str], optional): Dataset names if outputs shall be unlinked before. Defaults to None.
        remove (bool, optional): Shall output files automatically be deleted. Defaults to True.

    Raises:
        ParameterError: At least one of out_files, in_files or plan id need to be specified
        InputError: Could not find a plan associated to this id.
    """
    if out_files is None and in_files is None and plan_id is None:
        raise ParameterError(
            "Missing out_files, in_files and plan. \n"
            "Please specify at least one of out_files, in_files or plan to be reverted. \n"
        )

    activities = []
    out_fis = []
    project_context.clear()
    if in_files is not None:
        activities = flatten([Activity.filter(input=in_fi) for in_fi in in_files])

    if plan_id is not None:
        renku_plan = [plan for plan in Plan.list() if plan.id in plan_id]
        if len(renku_plan) == 0:
            raise InputError(
                f"Invalid plan_id. Could not find any plan associated with {plan_id} in this project."
            )
        activities = activities + renku_plan[0].activities

    if len(activities) > 0:
        out_fis = flatten(
            [[act_out.path for act_out in act.generated_outputs] for act in activities]
        )

    if out_files is not None:
        activities = activities + flatten(
            [Activity.filter(outputs=out_fi) for out_fi in out_files]
        )
        out_fis = out_fis + out_files

    if len(activities) == 0:
        print(
            f" WARNING: Could not find workflow executions related to your input:\n"
            f"out_files: {out_files}.\n"
            f"in_files: {in_files}.\n"
            f"plan_id: {plan_id}.\n"
            f"Nothing to revert."
        )
        return

    if dataset_name is not None:
        unlink_dataset_files(
            out_files=out_fis, dataset_name=dataset_name, remove=remove
        )

    # revert activities
    activity_ids = list(set([act.id for act in activities]))
    [
        omni_wflow.renku_workflow_revert(activity_id=act_id, plan=True)
        for act_id in activity_ids
    ]



import logging
import sys
import os
import functools


def get_logger(log_file_name, log_level=logging.INFO, log_sub_dir=""):
    """Creates a Log File and returns Logger object

    Args:
        log_level:     The log level.
        log_file_name: The log file name.
        log_sub_dir:   The directory to save log files to.
    """

    # Build Log File Full Path
    logPath = (
        log_file_name
        if os.path.exists(log_file_name)
        else os.path.join(log_sub_dir, (str(log_file_name) + ".log"))
    )

    # Create logger object and set the format for logging and other attributes
    logger = logging.Logger(log_file_name)

    # create formatter
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    # create file handler
    handler = logging.FileHandler(logPath, "a+")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # Set loglevel
    logger.setLevel(log_level)

    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    # add ch to logger
    logger.addHandler(ch)

    # Return logger object
    return logger


def log_decorator(_func=None):
    def log_decorator_info(func):
        @functools.wraps(func)
        def log_decorator_wrapper(self, *args, **kwargs):
            """Build logger object"""
            logger_obj = get_logger(
                log_file_name=self.log_file_name,
                log_level=self.log_level,
                log_sub_dir=self.log_file_dir,
            )

            """ Create a list of the positional arguments passed to function."""
            args_passed_in_function = [repr(a) for a in args]

            """ Create a list of the keyword arguments."""
            kwargs_passed_in_function = [f"{k}={v!r}" for k, v in kwargs.items()]

            """ The lists of positional and keyword arguments is joined together to form final string """
            formatted_arguments = ", ".join(
                args_passed_in_function + kwargs_passed_in_function
            )

            """log function begining"""
            logger_obj.info(f"Arguments: {formatted_arguments} - Begin function")
            try:
                """log return value from the function"""
                value = func(self, *args, **kwargs)
                logger_obj.info(f"Returned: - End function {value!r}")
            except Exception:
                """log exception if occurs in function"""
                logger_obj.error(f"Exception: {str(sys.exc_info()[1])}")
                raise
            return value

        return log_decorator_wrapper

    if _func is None:
        return log_decorator_info
    else:
        return log_decorator_info(_func)


def create_logger(log_level):
    """General logger for OmniBench class

    Args:
        log_level ([type]): The log level.
    """
    logger = logging.getLogger("OmniBench")
    logger.setLevel(log_level)

    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    # create formatter
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(ch)
    return logger



from omnibenchmark.management import general_checks
from omnibenchmark.utils.exceptions import InputError
import pytest
import requests

### Test is_renku_project
def test_is_renku_project_for_current_dir():
    assert not general_checks.is_renku_project()


def test_is_renku_project_for_non_exist_dir():
    with pytest.raises(Exception) as e_info:
        general_checks.is_renku_project("/non_existing/path")


### Test find_orchestrator
def test_find_orchestrator_matching_key(monkeypatch, mock_html_content):
    class MockResponse:
        def __init__(self):
            self.content = mock_html_content

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)
    assert (
        general_checks.find_orchestrator("omni_batch")
        == "https://renkulab.io/knowledge-graph/projects/omnibenchmark/orchestrator"
    )


def test_find_orchestrator_not_existing_header(monkeypatch, mock_html_content):
    class MockResponse:
        def __init__(self):
            self.content = mock_html_content

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    with pytest.raises(InputError, match=r"Could not find columns with names new*?"):
        general_checks.find_orchestrator("omni_batch", key_header="new")


def test_find_orchestrator_not_existing_key(monkeypatch, mock_html_content):
    class MockResponse:
        def __init__(self):
            self.content = mock_html_content

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)
    assert general_checks.find_orchestrator("omni") is None


@pytest.mark.api_call
def test_find_orchestrator_works():
    assert (
        general_checks.find_orchestrator("omni_batch_py")
        == "https://renkulab.io/knowledge-graph/projects/omnibenchmark/omni-batch-py/orchestrator-py"
    )



from omnibenchmark.management import wflow_checks
from renku.infrastructure.gateway.activity_gateway import ActivityGateway
from omnibenchmark.management import general_checks
import pytest

### Test activity_plan_is_valid
def test_activity_plan_is_valid_no_activity(act_gateway):

    assert isinstance(
        wflow_checks.activity_plan_is_valid(output="", activity_gateway=act_gateway),
        type(None),
    )


def test_activity_plan_is_valid_invalid_activity(
    mock_activity, monkeypatch, act_gateway
):

    mock_activity.association.plan.invalidated_at = "today"

    def get_mock_activity():
        return [mock_activity]

    monkeypatch.setattr(
        ActivityGateway,
        "get_activities_by_generation",
        lambda *args, **kwargs: get_mock_activity(),
    )

    assert isinstance(
        wflow_checks.activity_plan_is_valid(".", activity_gateway=act_gateway),
        type(None),
    )


def test_activity_plan_is_valid_valid_activity(mock_activity, monkeypatch, act_gateway):
    def get_mock_activity():
        return [mock_activity]

    def get_mock_status():
        return True

    monkeypatch.setattr(
        ActivityGateway,
        "get_activities_by_generation",
        lambda *args, **kwargs: get_mock_activity(),
    )
    monkeypatch.setattr(
        general_checks, "is_renku_project", lambda *args, **kwargs: get_mock_status()
    )

    assert (
        wflow_checks.activity_plan_is_valid("test_output", activity_gateway=act_gateway)
        == "test_output"
    )


### Test filter_activity_exist
def test_filter_activity_exist_gateway_establishment(mock_activity, monkeypatch):
    def get_mock_activity():
        return [mock_activity]

    monkeypatch.setattr(
        ActivityGateway,
        "get_activities_by_generation",
        lambda *args, **kwargs: get_mock_activity(),
    )

    def get_mock_status():
        return True

    monkeypatch.setattr(
        general_checks, "is_renku_project", lambda *args, **kwargs: get_mock_status()
    )

    assert wflow_checks.filter_activity_exist(["test_output1", "test_output2"]) == [
        "test_output1",
        "test_output2",
    ]


# Test get_all_plans
def test_get_all_plans_no_plans():
    assert wflow_checks.get_all_plans() == []


# Test filter_valid_plans
def test_filter_valid_plans_no_plans():
    assert wflow_checks.filter_valid_plans([]) == []


def test_filter_valid_plans_valid(mock_plan):
    assert wflow_checks.filter_valid_plans([mock_plan]) == [mock_plan]


def test_filter_valid_plans_valid_invalid(mock_plan):
    mock_plan.invalidated_at = "today"
    assert wflow_checks.filter_valid_plans([mock_plan]) == []


# Test get_plan_output_values
def test_get_plan_output_values_no_outputs(mock_plan):
    assert wflow_checks.get_plan_output_values(mock_plan) == []


def test_get_plan_output_values_detect_output(mock_plan, mock_output):
    mock_output.default_value = "path/to/default/output.test"
    mock_plan.outputs = [mock_output]
    assert wflow_checks.get_plan_output_values(mock_plan) == [
        "path/to/default/output.test"
    ]


# Test get_plan_id_from_output
def test_get_plan_id_from_output_without_plan(act_gateway):

    assert (
        wflow_checks.get_plan_id_from_output(
            "test_output", activity_gateway=act_gateway
        )
        == []
    )


def test_get_plan_id_from_output_with_plan(mock_activity, monkeypatch, act_gateway):
    def get_mock_activity():
        return [mock_activity]

    monkeypatch.setattr(
        ActivityGateway,
        "get_activities_by_generation",
        lambda *args, **kwargs: get_mock_activity(),
    )

    def get_mock_status():
        return True

    monkeypatch.setattr(
        general_checks, "is_renku_project", lambda *args, **kwargs: get_mock_status()
    )

    assert wflow_checks.get_plan_id_from_output(
        "test_output", activity_gateway=act_gateway
    ) == ["XXX"]


# Test find_plan_by_outputs
def test_find_plan_by_outputs_no_plan():
    assert wflow_checks.find_plan_by_outputs(["test_output"]) == []


def test_find_plan_by_outputs_with_plan(monkeypatch, mock_plan):
    def get_mock_plan():
        return [mock_plan]

    def mock_plan_id():
        return ["XXX"]

    monkeypatch.setattr(
        wflow_checks, "filter_valid_plans", lambda *args, **kwargs: get_mock_plan()
    )

    monkeypatch.setattr(
        wflow_checks, "get_plan_id_from_output", lambda *args, **kwargs: mock_plan_id()
    )

    assert wflow_checks.find_plan_by_outputs(["test_output"]) == [mock_plan]


# Test check_plan_exist
def test_check_plan_exist_no_plan():
    assert wflow_checks.check_plan_exist(["test_output1", "test/output2"]) == None


def test_check_plan_exist_one_plan(mock_plan, monkeypatch):
    def get_mock_plan():
        return [mock_plan]

    monkeypatch.setattr(
        wflow_checks, "find_plan_by_outputs", lambda *args, **kwargs: get_mock_plan()
    )

    assert wflow_checks.check_plan_exist(["test_output1", "test/output2"]) == mock_plan


def test_check_plan_exist_two_plans(mock_plan, monkeypatch):
    def get_mock_plan():
        return [mock_plan, mock_plan]

    monkeypatch.setattr(
        wflow_checks, "find_plan_by_outputs", lambda *args, **kwargs: get_mock_plan()
    )
    with pytest.raises(Exception) as e_info:
        wflow_checks.check_plan_exist(["test_output1", "test/output2"])



from renku.command.dataset import list_datasets_command
from renku.api import Dataset
import omnibenchmark.management.general_checks


def renku_dataset_list():
    if not omnibenchmark.management.general_checks.is_renku_project():
        return []
    result = list_datasets_command().build().execute()
    data_out = [data._subject for data in result.output]
    data_api = [Dataset._from_dataset(dat) for dat in data_out]
    return data_api



"""Checks concerning (usually renku) datasets"""

from renku.api import Activity, Project, Dataset
from omnibenchmark.renku_commands import renku_api
from omnibenchmark.utils.user_input_checks import flatten
import requests
from typing import Union, List, Mapping, Any
import os


def query_renku_api(
    url: str, page_num: int, page_item: int = 100
) -> List[Mapping[Any, Any]]:
    response = requests.get(url, params={"per_page": page_item, "page": page_num})
    return response.json()


def query_multipages(url: str, page_item: int = 100) -> List[Mapping[Any, Any]]:
    multi_page = True
    page_num = 1
    response: List = []
    while multi_page:
        res = query_renku_api(url, page_num=page_num, page_item=page_item)
        page_num += 1
        if len(res) < page_item:
            multi_page = False
        if isinstance(res, list):
            response.extend(res)
    return response


def renku_dataset_exist(name: str, path: Union[os.PathLike, str] = os.getcwd()) -> bool:
    """Check if a renku dataset with a specific name already exist at a certain project path.

    Args:
        name (str): Dataset name
        path (PathLike, optional): Project path to check. Defaults to ".".

    Returns:
        bool: True/False a dataset with that name exist in the project at that path.
    """

    current_dir = os.getcwd()
    os.chdir(path)
    datasets = renku_api.renku_dataset_list()
    # datasets = Dataset.list()
    matches = [dataset.name for dataset in datasets if dataset.name == name]
    os.chdir(current_dir)
    return True if len(matches) >= 1 else False


def dataset_name_exist(name: str, kg_url: str) -> bool:
    """Check if a renku dataset with a defined name already exists in the knowledge base.

    Args:
        name (str): Name to query
        kg_url (str): Url of the knowledge base to query

    Returns:
        bool: True/False, if a dataset with that name already exist
    """

    url = kg_url + "/datasets?query=" + name
    response = query_multipages(url)
    # Checks to ensure no complete name matching (Remove if file matching is moved to triplet store queries)
    if any(name in item["name"] for item in response) or any(
        item["name"] in name for item in response
    ):
        nl = "\n"
        name_list = set([item["name"] for item in response])
        print(
            f"A dataset with a complete match of {name} already exist.\n"
            "Please specify a name without a complete match within:\n"
            f"{nl}{nl.join(name_list)}"
        )
        return True
    return True if name in [item["name"] for item in response] else False


def find_activities_with_missing_inputs() -> List[Activity]:
    """Find activities related to missing input files

    Returns:
        List[Activity]: Activities that have missing input files linked
    """
    (
        outdated_generations,
        outdated_activities,
        modified_inputs,
        deleted_inputs,
    ) = Project().status()
    return Activity.filter(inputs=list(deleted_inputs))


def find_outputs_with_missing_inputs() -> List[str]:
    """Find Outputs that were generated by activities with missing inputs

    Returns:
        List[str]: List of output files, that have missing input files
    """
    act_list = find_activities_with_missing_inputs()
    out_list = flatten([act.generated_outputs for act in act_list])
    return list(set([out.path for out in out_list]))



from omnibenchmark.utils import benchmark_queries as omni
import pytest

# Test get_orchestrator_projects_from_cicd_yaml
@pytest.mark.api_call
def test_get_orchestrator_projects_from_cicd_yaml_existing_benchmark():
    o_url = "https://renkulab.io/knowledge-graph/projects/omnibenchmark/omni-batch-py/orchestrator-py"
    assert (
        "omnibenchmark/omni-batch-py/cellbench-py"
        in omni.get_orchestrator_projects_from_cicd_yaml(o_url)
    )


def test_get_orchestrator_projects_from_cicd_yaml_exclude_stage():
    o_url = "https://renkulab.io/knowledge-graph/projects/omnibenchmark/omni-batch-py/orchestrator-py"
    project_list = omni.get_orchestrator_projects_from_cicd_yaml(
        o_url, exclude_stages=["metric_run"]
    )
    assert "omnibenchmark/omni-batch-py/lisi-py" not in project_list
